unit NSHA256;

interface

type
  TSHA256 = class(TObject)
  private
    Hash: array[0..7] of Longword;
    Count: array[0..1] of Longword;
    Buffer: array[0..63] of Byte;
  public
    constructor Create;
    procedure Update(const Buf; NBytes: Cardinal);
    function Digest: string;
    function HexDigest: string;
  end;

implementation

const
  K: array[0..63] of Longword = (
    $428a2f98, $71374491, $b5c0fbcf, $e9b5dba5,
    $3956c25b, $59f111f1, $923f82a4, $ab1c5ed5,
    $d807aa98, $12835b01, $243185be, $550c7dc3,
    $72be5d74, $80deb1fe, $9bdc06a7, $c19bf174,
    $e49b69c1, $efbe4786, $0fc19dc6, $240ca1cc,
    $2de92c6f, $4a7484aa, $5cb0a9dc, $76f988da,
    $983e5152, $a831c66d, $b00327c8, $bf597fc7,
    $c6e00bf3, $d5a79147, $06ca6351, $14292967,
    $27b70a85, $2e1b2138, $4d2c6dfc, $53380d13,
    $650a7354, $766a0abb, $81c2c92e, $92722c85,
    $a2bfe8a1, $a81a664b, $c24b8b70, $c76c51a3,
    $d192e819, $d6990624, $f40e3585, $106aa070,
    $19a4c116, $1e376c08, $2748774c, $34b0bcb5,
    $391c0cb3, $4ed8aa4a, $5b9cca4f, $682e6ff3,
    $748f82ee, $78a5636f, $84c87814, $8cc70208,
    $90befffa, $a4506ceb, $bef9a3f7, $c67178f2
    );


function Ch(X, Y, Z: Longword): Longword;
begin
  Result := X and (Y xor Z) xor Z;
end;

function Maj(X, Y, Z: Longword): Longword;
begin
  Result := X and (Y or Z) or (Y and Z)
end;

function RotR(X, N: Longword): Longword;
begin
  Result := (X shr N) or (X shl (32-N));
end;

function S0(X: Longword): Longword;
begin
  Result := RotR(X, 2) xor RotR(X, 13) xor RotR(X, 22);
end;

function S1(X: Longword): Longword;
begin
  Result := RotR(X, 6) xor RotR(X, 11) xor RotR(X, 25);
end;

function Ss0(X: Longword): Longword;
begin
  Result := RotR(X, 7) xor RotR(X, 18) xor (X shr 3);
end;

function Ss1(X: Longword): Longword;
begin
  Result := RotR(X, 17) xor RotR(X, 19) xor (X shr 10);
end;

procedure EncodeBE(const InBuf: array of Longword; var OutBuf);
var
  i: Cardinal;
begin
  for i := 0 to High(InBuf) do
  begin
    PChar(@OutBuf)[i*4] := Char(InBuf[i] shr 24 and $ff);
    PChar(@OutBuf)[i*4+1] := Char(InBuf[i] shr 16 and $ff);
    PChar(@OutBuf)[i*4+2] := Char(InBuf[i] shr 8 and $ff);
    PChar(@OutBuf)[i*4+3] := Char(InBuf[i] and $ff);
  end;
end;

procedure DecodeBE(const InBuf; InBytes: Cardinal;
  var OutBuf: array of Longword);
var
  i: Cardinal;
begin
  for i := 0 to (InBytes div 4)-1 do
    OutBuf[i] := Longword(PChar(@InBuf)[i*4]) shl 24 or
                 Longword(PChar(@InBuf)[i*4+1]) shl 16 or
                 Longword(PChar(@InBuf)[i*4+2]) shl 8 or
                 Longword(PChar(@InBuf)[i*4+3]);
end;

procedure Transform(const Block; var Hash: array of Longword);
var
  a, b, c, d, e, f, g, h, t1, t2: Longword;
  w: array[0..63] of Longword;
  i: Integer;
begin
  DecodeBE(Block, 64, w);
  for i := 16 to 63 do
    w[i] := Ss1(w[i-2]) + w[i-7] + Ss0(w[i-15]) + w[i-16];

  a := Hash[0];
  b := Hash[1];
  c := Hash[2];
  d := Hash[3];
  e := Hash[4];
  f := Hash[5];
  g := Hash[6];
  h := Hash[7];

  for i := 0 to 63 do
  begin
    t1 := h + S1(e) + Ch(e, f, g) + K[i] + w[i];
    t2 := S0(a) + Maj(a, b, c);
    h := g;
    g := f;
    f := e;
    e := d + t1;
    d := c;
    c := b;
    b := a;
    a := t1 + t2;
  end;

  Inc(Hash[0], a);
  Inc(Hash[1], b);
  Inc(Hash[2], c);
  Inc(Hash[3], d);
  Inc(Hash[4], e);
  Inc(Hash[5], f);
  Inc(Hash[6], g);
  Inc(Hash[7], h);
end;

constructor TSHA256.Create;
begin
  Hash[0] := $6a09e667;
  Hash[1] := $bb67ae85;
  Hash[2] := $3c6ef372;
  Hash[3] := $a54ff53a;
  Hash[4] := $510e527f;
  Hash[5] := $9b05688c;
  Hash[6] := $1f83d9ab;
  Hash[7] := $5be0cd19;
end;

procedure TSHA256.Update(const Buf; NBytes: Cardinal);
var
  i, index, partlen, nbits: Cardinal;
begin
  index := Count[1] shr 3 and $3F;
  nbits := NBytes shl 3;
  Inc(Count[1], nbits);
  if Count[1] < nbits then
    Inc(Count[0]);
  Inc(Count[0], NBytes shr 29);

  partlen := 64 - index;
  if NBytes >= partlen then
  begin
    Move(Buf, Buffer[index], partlen);
    Transform(Buffer, Hash);
    i := partlen;
    while (i + 63) < NBytes do
    begin
      Transform(PChar(@Buf)[i], Hash);
      Inc(i, 64);
    end;
    index := 0;
  end else
    i := 0;
  Move(PChar(@Buf)[i], Buffer[index], NBytes-i);
end;

function TSHA256.Digest: string;
const
  padding: array[0..63] of Byte = (
    $80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  );
var
  bits: array[0..7] of Char;
  index, padlen: Cardinal;
  hash: array[0..7] of Longword;
  count: array[0..1] of Longword;
begin
  Move(Self.Hash, hash, SizeOf(hash));
  Move(Self.Count, count, SizeOf(count));
  EncodeBE(Count, bits);
  index := (Count[1] shr 3) and $3f;
  if index < 56 then
    padlen := (56 - index)
  else
    padlen := (120 - index);
  Update(padding, padlen);
  Update(bits, 8);
  SetString(Result, nil, 32);
  EncodeBE(Self.Hash, Result[1]);
  Move(hash, Self.Hash, SizeOf(Self.Hash));
  Move(count, Self.Count, SizeOf(Self.Count));
end;

function TSHA256.HexDigest: string;
const
  hexdigs: array[0..15] of Char = '0123456789abcdef';
var
  i: Integer;
  s: string;
begin
  s := Digest;
  SetString(Result, nil, Length(s)*2);
  for i := 1 to Length(s) do
  begin
    Result[(i-1)*2+1] := hexdigs[Byte(s[i]) shr 4];
    Result[(i-1)*2+2] := hexdigs[Byte(s[i]) and $F];
  end;
end;

end.
